"""
This class implements a buffer containing a set of states. It will serve as interface with the
individual states.

Author: Eric Canas.
Github: https://github.com/Eric-Canas
Email: eric@ericcanas.com
Date: 16-07-2022
"""
from __future__ import annotations
from collections import deque

import numpy as np
from loguru import logger
from vid2info.inference.config import TRACK_INFO, TRACK_ID, CLASS_NAME

from vid2info.state.finite_state_machine.track_recoverer import TrackRecoverer
from vid2info.utils.config import TIME_FORMAT, ELEMENT_AGE_TIME_FORMAT
from vid2info.utils.general import time_as_str
from vid2info.state.config import ELEMENT_BUFFER_SIZE, INITIAL_TIMESTAMP, OUT_TIMESTAMP, INITIAL_CLASS_NAME, \
    OUT_CLASS_NAME, FINITE_STATE_MACHINE, TRACK_LENGTH, INITIAL_BBOX_XYXY, OUT_BBOX_XYXY
from vid2info.state.finite_state_machine.config_keys import TIMESTAMP, PREVIOUS_STATE, NEW_STATE
from vid2info.state.finite_state_machine.element_finite_state_machine import ElementFiniteStateMachine
from vid2info.state.config import BUFFER_SIZE
from vid2info.state.scene_state import SceneState
from vid2info.state.element_state.element_state import ElementState
from time import time


class StateBuffer:
    def __init__(self, buffer_size : int = BUFFER_SIZE, running_online : bool = False,
                 scene_state_class : callable = SceneState, element_state_class : callable = ElementState,
                 finite_state_machines_config: dict | None = None, keep_elements_history: bool = False,
                 track_recoverer: TrackRecoverer | None = None):
        """
        Initialize the StateBuffer. It is used for tracking the history of the last scene and
         element states.

        :param buffer_size: int. The size of the history buffer to keep.
        :param running_online: bool. If True, it will assume that the video is comming from an online stream
                like a webcam. So the timestamps of the states will update with the current time in every call.
                If False, it will assume that the video is coming from a pre-recorded file. So the timestamps
                will have to be served from the video.
        :param scene_state_class: type. The class of the scene state. It should be a subclass of SceneState.
        :param element_state_class: Class. The class of the element states. It should be a subclass of ElementState.
        :param finite_state_machines_config: dict. The configuration of the finite state machines for the elements.
                It is a dictionary with the track_id as key and the configuration as value.
        :param keep_elements_history: bool. If True, the element states will be kept in the buffer. If False, the
                element states will be removed from the buffer when they are not in the scene anymore.
        :param recover_tracks_for_static_elements: bool. Only used when finite_state_machines_config is passed.
                If True, it will try to recover the tracks for the elements with a class that fallbacks in a default
                state machine. NOTE: It should ONLY be used when we are expecting very STATIC elements that do not tend
                to overlap with each other. Otherwise, it could produce unexpected results. The expected scenario for
                this flag is when we do expect static elements, that can be partially lost by occlusions, and that have
                an attached state machine that makes transitions based in the object class but that only accept a few
                 entry classes but the classes. For example: A lot of robotic faces that always starts with the class
                 "neutral expression" and for which we are registering the expression transitions.
        :param recover_track_attempts: int. Only used when finite_state_machines_config is passed and
                recover_tracks_for_static_elements is True. The number of attempts to recover the track of the elements
                that fallback in a default state machine.
        """
        self.buffer_size = max(buffer_size, 1)
        self.buffer = deque(maxlen=self.buffer_size)
        self.running_online = running_online
        self.scene_state_class = scene_state_class
        self.element_state_class = element_state_class
        self.finite_state_machines_config = finite_state_machines_config
        self.keep_elements_history = keep_elements_history
        self.elements_history = {}
        assert track_recoverer is None, f"Track recoverer is momentarily disabled."
        self.track_recoverer = track_recoverer
        if self.track_recoverer is not None:
            self.track_recoverer.state_buffer = self

        self.recovered_tracks_correspondences = {}

    def build_scene_state(self, track_info: dict, add_to_buffer : bool = True,
                          frame: np.ndarray | None = None) -> SceneState:
        """
        Create a new state from the tracker output.

        :param track_info: dictionary. The output of the tracker. It is the dictionary outputted by
                get_track_ids when return_as_dict is True. It's keys are the track_ids and the values are
                dictionaries with, at least, the following keys: ('bbox_xyxy', 'confidence', 'class_id', 'track_length')
        :param add_to_buffer: bool. If True, the state will be added to the buffer.
        :param frame: np.ndarray or None. The frame of the scene. If given, it will save a cropped subimage for
                each detected element.
        :param segmentation_mask: np.ndarray or None. The segmentation mask of the scene. If given, it will save a
                cropped segmentation mask for each detected element.
        :return: SceneState. The new state. If add_to_buffer was True it can be also found in the last element of the
                current self.buffer.
        """
        scene_state = self.scene_state_class(track_info=track_info, buffer=self,
                                             frame=frame,
                                             element_state_class=self.element_state_class,
                                             element_state_machine_configs=self.finite_state_machines_config)
        if add_to_buffer:
            self.buffer.append(scene_state)
            if self.keep_elements_history:
                self.update_history(scene_state)
        return scene_state


    def get_first_element_timestamp(self, track_id : int) -> float | None:
        """
        Get the first_timestamp of the first detection of the element with the given track_id in the buffer
        or current_time if the element is not in the buffer, and it is running online (webcam).
        If the element is not in the buffer and is not running online, it will return None.

        :param track_id: int. The track_id of the element to get the first_timestamp of.

        :return: float or None. The first_timestamp of the first detection of the element with the given track_id
                in the buffer or current_time if the element is not in the buffer, and it is running online (webcam).
                If the element is not in the buffer and is not running online, it will return None.
        """
        for scene_state in self.buffer:
            if track_id in scene_state.elements:
                track_id = self.recovered_tracks_correspondences.get(track_id, track_id)
                return scene_state.elements[track_id].first_detection_timestamp
        if self.running_online:
            return time()
        else:
            raise NotImplementedError("Not yet implemented for the case where the video is not coming from an online stream.")


    def get_last_detection_of_element(self, track_id: int) -> ElementState | None:
        """
        Get the last detection of the element with the given track_id in the buffer or None if
        the element is not in the buffer.

        :param track_id: int. The track_id of the element to get the last detection of.

        :return: ElementState or None. The last detection of the element with the given track_id in the buffer or None if
                the element is not in the buffer.
        """
        for scene_state in reversed(self.buffer):
            if track_id in scene_state.elements:
                return scene_state.elements[track_id]
        return None

    def get_element_buffer(self, track_id: int, buffer_size: int = ELEMENT_BUFFER_SIZE) -> deque:
        """
        Returns a buffer with the last buffer_size elements with the given track_id in the scene_buffer. If the track_id
        disappears in the scene_buffer for some in between frames, these frames will be ignored. So the buffer will be
        not representative of the distance between frames.

        :param track_id: int. The track_id of the element we want to get the buffer for.
        :param buffer_size: int. The maximum size of the expected buffer.

        :return deque. A buffer containing the last buffer_size elements with the given track_id in the scene_buffer.
        """
        element_buffer = deque(maxlen=buffer_size)
        # Read the scene_buffer backwards
        for scene_state in reversed(self.buffer):
            if track_id in scene_state.elements:
                element_buffer.append(scene_state.elements[track_id])
            if len(element_buffer) == buffer_size:
                break
        element_buffer = deque(reversed(element_buffer))
        return element_buffer

    def get_element_state_machine(self, element_info: dict, in_scene_tracks: tuple | list = ()) -> ElementFiniteStateMachine:
        """
        Returns the state machine of the element with the given track_id. If this element is new,
        it will create a new state machine using the configs given in element_state_machine_configs.

        :param track_id: int. The track_id of the element.
        :param current_class : str | None. The current class of the element, only used if the element is new.
        :param bbox_xyxy: np.ndarray | None. The bounding box of the element, only used if the element is new.
        :param recover_default_machines: bool. If True, the default state machines will be recovered if the element is new.
        :param in_scene_tracks: tuple | list. The tracks of the elements that are actually in the scene. Only used if
                recover_default_machines is True.

        :return ElementFiniteStateMachine. The current state machine of the element.
        """
        # Check for the last aparition of the element in the buffer.
        finite_state_machine = None
        current_class, track_id = element_info[CLASS_NAME], element_info[TRACK_INFO][TRACK_ID]
        last_detection = self.get_last_detection_of_element(track_id=track_id)
        if last_detection is not None:
            finite_state_machine = last_detection.finite_state_machine
        if finite_state_machine is None:
            if self.track_recoverer is not None:
                # It will try to find the old track_id for the element. Filling its correspondences buffer (lately used
                # by the SceneState) if it is able to find one. Otherwise, it will return None or create a new one.
                finite_state_machine = self.track_recoverer.recover_track_id_from_state_machine(
                    last_detection=last_detection, track_id=track_id, current_class=current_class, bbox_xyxy=bbox_xyxy,
                    recover_default_machines=True)
            else:

                if current_class not in self.finite_state_machines_config:
                    assert 'default' in self.finite_state_machines_config, f"Element class {current_class} not found " \
                                                                           f"in finite state machine configs and no default " \
                                                                           f"config found."
                    current_class = 'default'
                config = self.finite_state_machines_config[current_class]
                finite_state_machine = ElementFiniteStateMachine(config=config)
        return finite_state_machine


    def update_history(self, state : SceneState):
        """
        Update the elements_history dictionary. It will keep the track of all the elements that have been
        seen during the analysis.
        """
        for track_id, element in state.elements.items():
            if self.track_recoverer is not None and \
                    self.track_recoverer.track_id_is_being_recovered(track_id=track_id, state_machine=element.finite_state_machine):
                # This way, the track_id won't be added to the history if it is being recovered.
                continue
            if track_id not in self.elements_history:
                initial_hist_state = {
                    INITIAL_TIMESTAMP: element.first_detection_timestamp,
                    INITIAL_CLASS_NAME: element.class_name,
                    FINITE_STATE_MACHINE: element.finite_state_machine,
                    INITIAL_BBOX_XYXY: element.bbox_xyxy,
                }
                self.elements_history[track_id] = initial_hist_state

            element_hist_update = {
                OUT_TIMESTAMP: element.timestamp,
                OUT_CLASS_NAME: element.class_name,
                OUT_BBOX_XYXY: element.bbox_xyxy,
                TRACK_LENGTH: element.track_length
            }
            self.elements_history[track_id].update(element_hist_update)

    def print_elements_history(self, min_track_length: int = 0):
        """
        Print the elements_history dictionary.

        :param min_track_length: int. The minimum track length to print it.
        """
        assert self.keep_elements_history, "The elements_history dictionary is not kept. Impossible to print it."

        for track_id, element in self.elements_history.items():
            track_len = element[TRACK_LENGTH]
            if track_len > min_track_length:
                start, end = element[INITIAL_TIMESTAMP], element[OUT_TIMESTAMP]
                print("Track id: ", track_id)
                print(f"\t From: {time_as_str(start, time_format=TIME_FORMAT)} - To: {time_as_str(end, time_format=TIME_FORMAT)} "
                      f"(Age: {time_as_str(end - start, time_format=ELEMENT_AGE_TIME_FORMAT)})")
                print(f"\t Entered as: {element[INITIAL_CLASS_NAME]} - Left as: {element[OUT_CLASS_NAME]}")
                print(f"\t Track length: {track_len}")
                print("\t Last known bbox: ", tuple(round(coord, ndigits=2) for coord in element[OUT_BBOX_XYXY]))
                print(f"\t Finite state machine: \n{element[FINITE_STATE_MACHINE]}")

    def get_elements_time_per_state(self, class_to_state_dict: dict = {})-> dict[int, dict[str, float]]:
        """
        Returns a dictionary with the time spent in each state for each element.

        :return dict[int, dict[str, float]]. A dictionary with the time spent in each state for each element.
        """
        assert self.keep_elements_history, "The elements_history dictionary is not kept. Impossible to print it."
        elements_time_per_state = {}
        for track_id, element in self.elements_history.items():
            element_info = {}
            initial_timestamp, history = element[INITIAL_TIMESTAMP], element[FINITE_STATE_MACHINE].history
            initial_class = class_to_state_dict.get(element[INITIAL_CLASS_NAME], element[INITIAL_CLASS_NAME])
            if len(history) == 0:
                out_class_name = class_to_state_dict.get(element[OUT_CLASS_NAME], element[OUT_CLASS_NAME])
                if out_class_name != initial_class:
                    logger.warning(f"Element {track_id} has no history. That's a missing track.")
                initial_time = element[OUT_TIMESTAMP] - initial_timestamp
                element_info[element[INITIAL_CLASS_NAME]] = initial_time
            else:
                previous_timestamp = initial_timestamp
                assert history[0][PREVIOUS_STATE] == initial_class, "The first state of the element is not the initial class."
                for entry in history:
                    change_timestamp, lapse_state = entry[TIMESTAMP], entry[PREVIOUS_STATE]
                    element_info[lapse_state] = change_timestamp - previous_timestamp
                    previous_timestamp = change_timestamp
                lapse_state = class_to_state_dict.get(element[OUT_CLASS_NAME], element[OUT_CLASS_NAME])
                if history[-1][NEW_STATE] != lapse_state:
                    logger.warning(f"The last state of the element is not the final class. Probably the track_id {track_id} has been lost")
                    continue
                element_info[lapse_state] = element[OUT_TIMESTAMP] - previous_timestamp
            elements_time_per_state[track_id] = element_info

        return elements_time_per_state



"""
This class implements the state of a concrete frame.

Author: Eric Canas.
Github: https://github.com/Eric-Canas
Email: eric@ericcanas.com
Date: 17-07-2022
"""
from __future__ import annotations
from time import time
import numpy as np

from vid2info.inference.utils import crop_bbox_from_image
from vid2info.inference.config import BBOX_XYXY, CLASS_NAME, CLASS_ID
from vid2info.state.element_state.element_state import ElementState

crop_img_if_given = lambda img, bbox, copy = True: crop_bbox_from_image(image=img, bbox_xyxy=bbox, is_normalized=True, copy=copy) \
                                        if img is not None else None

class SceneState:
    def __init__(self, track_info : dict, buffer, get_first_element_time_stamp: callable = lambda track_id : None,
                 frame: np.ndarray | None = None, segmentation_mask : dict | None = None,
                 save_full_segmentation_mask : bool = False, element_state_class : callable = ElementState,
                 element_state_machine_configs : dict | None = None):
        """
        Initialize the SceneState. It is used for defining the current scene, and will contain the
        state of each individual element in the scene.

        :param track_info: dictionary. The output of the tracker. It is the dictionary outputted by
                get_track_ids when return_as_dict is True. It's keys are the track_ids and the values are
                dictionaries with, at least, the following keys: ('bbox_xyxy', 'confidence', 'class_id', 'track_length')
        :param buffer: StateBuffer. The StateBuffer containing the N previous SceneStates.
        :param get_first_element_time_stamp: callable. A function that returns the first_timestamp of the first
                detection of the element with the given track_id in the buffer or the current timestamp if the
                element is not in the buffer.
        :param frame: np.ndarray or None. The frame of the scene. If given, it will save a cropped subimage for
                each detected element.
        :param segmentation_mask: dict or None. The segmentation mask of the scene, embedded within a dictionary
                with the following keys: ('segmentation_mask', 'background_class_idx', 'class_names_list'). If given,
                it will save a cropped subimage for each detected element.

        :param save_full_segmentation_mask: bool. If True, it will save the full segmentation mask of the scene.
        """

        self.elements = {}
        in_scene_tracks = tuple(track_info.keys())
        for track_id, element in track_info.items():
            finite_state_machine = None
            if element_state_machine_configs is not None:
                finite_state_machine = buffer.get_element_state_machine(element_info=element, in_scene_tracks=in_scene_tracks)

            self.elements[track_id] = element_state_class(element_tracker_info=element,
                                             first_detection_timestamp=buffer.get_first_element_timestamp(track_id=track_id),
                                             element_buffer=buffer.get_element_buffer(track_id=track_id),
                                             finite_state_machine=finite_state_machine, is_occluded=self.is_element_occluded(track_id=track_id, all_elements=track_info))

        self.timestamp = time()

    def __len__(self):
        return len(self.elements)

    @property
    def older_element_timestamp(self):
        return min(element.timestamp for element in self.elements.values())

    def get_elements_age_in_seconds(self) -> dict:
        """
        Returns the age of each element in the scene in seconds.

        :return dict. The age of each element in the scene. The keys are the track_ids and
                        the values are the age in seconds.
        """
        return {track_id : self.elements[track_id].age_in_seconds for track_id in self.elements}

    def get_element_counts(self) -> dict:
        """
        Returns the number of elements in the scene for each class.

            Returns:
                dict. The number of elements in the scene for each class. The keys are the class names and
                the values are the number of elements in the scene for that class.
            """
        classes_list = [element.class_name for element in self.elements.values()]
        return {class_name: classes_list.count(class_name) for class_name in set(classes_list)}

    def elements_per_state(self) -> dict[str, set[int]]:
        """
        Returns the elements of the scene in each state.

        :return dict[str, set[int]]. The elements of the scene in each state. The keys are the state names and
                the values are the set of track_ids of the elements in that state.
        """
        states = {}
        for _id, element in self.elements.items():
            state_machine = element.finite_state_machine
            if state_machine is not None:
                state = state_machine.current_state
                if state not in states:
                    states[state] = set()
                states[state].add(_id)
        return states

    def is_element_occluded(self, track_id, all_elements):
        # TODO: Implement it.
        return False



from ultralytics import YOLO

# Load a pretrained YOLO model (recommended for training)
model = YOLO('C:\\Code\\vid2info\\vid2info\\inference\\models\\segmentation\\yolov8n-seg.pt')


# Train the model using the 'coco128.yaml' dataset for 3 epochs
results = model.train(data='C:\\Code\\vid2info\\training_tools\\datasets\\instance-cooking-segmentation\\images\\val', epochs=3)

print(results)


"""
This class implements a calculator that computes all the information related to the movement of a tracked element.

Author: Eric Canas.
Github: https://github.com/Eric-Canas
Email: eric@ericcanas.com
Date: 27-02-2022
"""

from __future__ import annotations

from collections import deque
import itertools
import numpy as np

from vid2info.state.element_state.utils import frames_to_past_from_seconds_in_element_buffer
from vid2info.state.element_state.constants import STD_OUTLIER_DEVIATION, MIN_POINTS_TO_CALCULATE_OUTLIERS, \
    TOP, RIGHT, BOTTOM, LEFT


class ElementMovement:
    def __init__(self, element_state: 'ElementState', element_buffer: deque,
                 std_outlier_deviation: float = STD_OUTLIER_DEVIATION):
        """
        Initialize the ElementMovement calculator.

        :param element_state: ElementState. The element state.
        :param element_buffer: deque. The buffer containing the N previous ElementStates. It can help to compute the
                element evolution.
        """
        #TODO: Movement should be calculated in pixel space, not in percentage of the image. As it is messing the aspect ratio
        self.movement_vector_xy_1_sec = self.get_movement_vector(element_state=element_state, element_buffer=element_buffer, window_size_seconds=1)
        self.movement_1_sec_changed_significantly_xy = self.is_movement_vector_significant(element_buffer=element_buffer, movement_vector_variable='movement_vector_xy_1_sec',
                                                                                           deviation_from_std_to_be_relevant=std_outlier_deviation)
        self.movement_1_sec_changed_significantly = any(self.movement_1_sec_changed_significantly_xy)
        self.movement_1_sec_direction = self.get_movement_direction(movement_vector=self.movement_vector_xy_1_sec)
        # TODO: Could be interesting to use the area increment or decrement to infer the movement vector in Z axis.

    def get_movement_vector(self, element_state: 'ElementState', element_buffer: deque, window_size_seconds: int = 1) -> tuple[float, float]:
        """
        Compute the movement vector of the element.

        :param element_state: ElementState. The current state of the element.
        :param element_buffer: deque. The buffer containing the N previous ElementStates. It will help to compute the
                element evolution.
        :param window_size_seconds: int. The size of the window of frames to compute the movement vector.

        :return: tuple[float, float]. The movement vector of the element.
        """
        # Find the window size in frames
        curr_element_time = element_state.timestamp
        if len(element_buffer) == 0:
            return (0., 0.)

        windows_size_frames = frames_to_past_from_seconds_in_element_buffer(element_buffer=element_buffer,
                                                                            current_element_timestamp=curr_element_time,
                                                                            seconds_to_check_into_the_past=window_size_seconds)
        # Calculate the new immediate movement vector
        (curr_x, curr_y), (prev_x, prev_y) = element_state.cxcyn, element_buffer[-windows_size_frames].cxcyn
        return (prev_x - curr_x, prev_y - curr_y)

    def is_movement_vector_significant(self, element_buffer: deque, movement_vector_variable: str,
                                       seconds_to_check_into_the_past: int|None = None,
                                       deviation_from_std_to_be_relevant: float = STD_OUTLIER_DEVIATION,
                                       min_window_size_to_calculate: int = MIN_POINTS_TO_CALCULATE_OUTLIERS) -> tuple[bool, bool]:
        """
        Check if the movement vector of the element is significant.

        :param element_buffer: deque. The buffer containing the N previous ElementStates. It will help to compute the
                element evolution.
        :param movement_vector_variable: str. The name of the movement vector variable to check.
        :param seconds_to_check_into_the_past: int|None. The number of seconds to check into the past. If None, it will
                check the whole buffer.

        :return: tuple[bool, bool]. A tuple of booleans indicating if the movement vector is significant by axis.
        """
        if len(element_buffer) < min_window_size_to_calculate:
            return False, False
        windows_size_frames = frames_to_past_from_seconds_in_element_buffer(element_buffer=element_buffer,
                                                                            current_element_timestamp=element_buffer[-1].timestamp,
                                                                            seconds_to_check_into_the_past=seconds_to_check_into_the_past)
        # It is negative (from the future to the past) so we need to invert it
        window_size_beginning = len(element_buffer) - windows_size_frames

        assert hasattr(self, movement_vector_variable), f'Variable {movement_vector_variable} not found in ElementMovement'
        movement_vector = getattr(self, movement_vector_variable)
        assert len(movement_vector) == 2, f'Variable {movement_vector_variable} must be a tuple of length 2 (x, y)'
        # Get all the movement vectors of the last window_size_frames
        prev_movement_vectors = [getattr(prev_element.movement, movement_vector_variable)
                                    for prev_element in itertools.islice(element_buffer, window_size_beginning, None)]
        # Compute the standard deviation of the movement vectors
        mean, std = np.mean(prev_movement_vectors, axis=0), np.std(prev_movement_vectors, axis=0)
        # Check if the movement vector is significant (by axis)
        relevance_x, relevance_y = np.abs(mean - movement_vector) > deviation_from_std_to_be_relevant * std
        return bool(relevance_x), bool(relevance_y)

    def get_movement_direction(self, movement_vector: tuple[float, float]) -> str|None:
        """
        Get the direction of the movement vector among top, right, bottom & left (or None if (0., 0.)).

        :param movement_vector: tuple[float, float]. The movement vector.

        :return: str|None. The direction of the movement vector (or None if (0., 0.)).
        """
        assert len(movement_vector) == 2, f'Movement vector must be a tuple of length 2 (x, y)'
        x, y = movement_vector
        # Check if x or y are close to 0
        if abs(x) < 1e-5 and abs(y) < 1e-5:
            return None
        elif abs(x) > abs(y):
            return RIGHT if x > 0 else LEFT
        else:
            return BOTTOM if y > 0 else TOP




"""
This class implements the YoloDetector. It is used to detect elements in a
given image or batch of images. It uses YOLOv6 as the backbone of the detector.

Author: Eric Canas.
Github: https://github.com/Eric-Canas
Email: eric@ericcanas.com
Date: 09-07-2022
"""
from __future__ import annotations
import numpy as np

import os
import torch
import torch.nn.functional as F
import yaml

from vid2info.inference.segmentation.config import WEIGTHS_PATH, PARAMS_YAML, \
    MODEL_NAME, PIDNET_SCALE_FACTOR
from vid2info.inference.config import INFERENCE_DEVICE, CLASS_NAMES_LIST, SEGMENTATION_MASK, BACKGROUND_CLASS_IDX
from vid2info.inference.utils import image_to_float_tensor, get_l_r_pad

from third_party.pidnet.pidnet import get_pred_model

UINT8_TO_FLOAT = 1/255

class PidNetSegmenter:
    def __init__(self, weights : str = WEIGTHS_PATH, params_yml : str = PARAMS_YAML,
                 model_name = MODEL_NAME, inference_device : str = INFERENCE_DEVICE):
        """
        Initialize the segmenter. Load the model and the parameters.

        :param weights: str. Path to the weights file. It must be the .pt file generated during
                        the training of PIDNet.
        :param params_yml: str. Path to the yaml file with the model parameters. It must contain the following
                             parameters:
                                - mean : list of 3 elements. The channel-wise mean of the dataset. Set on the training
                                        to normalize the input.
                                - std : list of 3 elements. The channel-wise standard deviation of the dataset. Set on
                                        the training to normalize the input.
                                - class_names : list of strings. The class names of the dataset.
                                - background_class : string. The name of the background class.
        :param model_name: str ending with small, medium or large. Ex: pidnet_small. The version of the
                        pidnet model to load.
        :param inference_device: str. The device to use for inference. Must be 'cpu', 'cuda' or 'cuda:i'.
        """
        assert os.path.isfile(weights), f'Weights file {weights} not found.'
        if inference_device.startswith('cuda'):
            assert torch.cuda.is_available(), 'CUDA is not available.'
        else:
            assert inference_device == 'cpu', f'Inference device must be cpu or cuda. Got {inference_device}.'

        self.scale_factor = PIDNET_SCALE_FACTOR

        self.inference_device = torch.device(inference_device)

        self.load_params(params_yaml=params_yml)

        self.model = self.build_model(model_name=model_name, weights=weights).to(self.inference_device)
        # Warm the model up (Don't set in self to let the garbage collector free the memory)
        input_sample = torch.zeros(1, len('RGB'), self.scale_factor*2, self.scale_factor*2, dtype=torch.float32,
                                   device=self.inference_device)
        self.model(input_sample)

    def load_params(self, params_yaml : str = PARAMS_YAML):
        """
        Load the parameters from the given yml file. That params are mean, std and class_names.

        :param params_yaml: str. Path to the yaml file.
        """
        assert os.path.isfile(params_yaml), f'Dataset yml file {params_yaml} not found.'

        with open(params_yaml, 'r') as f:
            params = yaml.load(f, Loader=yaml.FullLoader)

        self.class_names = tuple(params['class_names'])
        self.num_classes = len(self.class_names)
        self.background_class = params['background_class']
        self.background_class_idx = self.class_names.index(self.background_class)
        assert self.background_class_idx >= 0, f'Background class {self.background_class} not found in class names.'

        self.mean = torch.tensor(data=params['mean'], dtype=torch.float32, device=self.inference_device)
        self.std = torch.tensor(data=params['std'], dtype=torch.float32, device=self.inference_device)

        assert len(self.mean) == len(self.std) == 3, f'Mean and std must be a list of 3 elements (RGB). ' \
                                                     f'Got {len(self.mean)} and {len(self.std)}.'
        assert self.num_classes > 1, f'There must be at least 2 classes to segment. Got {self.num_classes}.'

    def build_model(self, model_name : str = MODEL_NAME, weights : str = WEIGTHS_PATH) -> torch.nn.Module:
        """
        Build the model, loading the given weights.

        :param model_name: str ending with small, medium or large. Ex: pidnet_small. The version of the
                        pidnet model to load.
        :param weights: str. Path to the weights file.
        :param inference_device: str. The device to use for inference. Must be 'cpu' or 'cuda'.
        """
        assert os.path.isfile(weights), f'Weights file {weights} not found.'
        # Get the model
        model = get_pred_model(name=model_name, num_classes=self.num_classes).to('cpu')
        # Load the weights
        state_dict = torch.load(f=weights, map_location='cpu')
        state_dict = state_dict['state_dict'] if 'state_dict' in state_dict else state_dict

        # Adapt the weights to the prediction model
        layers_offset = 'model.'
        model_dict = model.state_dict()
        layers_have_offset = all(k.startswith(layers_offset) for k in state_dict.keys() if not k.endswith('criterion.weight'))
        name_offset = len(layers_offset) if layers_have_offset else 0

        state_dict = {layer_name[name_offset:]: state for layer_name, state in state_dict.items()
                      if (layer_name[name_offset:] in model_dict and
                          state.shape == model_dict[layer_name[name_offset:]].shape)}
        assert len(state_dict) == len(model_dict), f'Weights file {weights} has different layers than the model.' \
                                                   f'Got {len(state_dict)} layers, expected {len(model_dict)}.' \
                                                   f'Are you sure you are using the correct weights file?'
        # Load them to the model
        model.load_state_dict(state_dict=state_dict, strict=True)
        model.eval()

        return model

    def segment(self, image: np.ndarray | torch.Tensor, is_bgr : bool = False,
                prediction_as_tensor : bool = False, return_as_dict: bool = True) -> np.ndarray | torch.Tensor | dict:
        """
        Infer the segmentation map of the given image

        :param image: np.ndarray or torch.tensor as uint8 (0, 255) or float32 (0., 1.). The image to segment,
                      in format (H, W, C). Channels must be 3, RGB or BGR.
        :param is_bgr: bool. If True, the image is in BGR format. If False, the image is in RGB format.
        :param prediction_as_tensor: bool. If True, the output is a torch.tensor. If False, the output is a np.ndarray.
        :param return_as_dict: bool. If True, the output is a dict containing the segmenter meta-information.
                               If False, the output is directly the prediction.

        :return: np.ndarray or torch.Tensor in format uint8. The segmentation map, in format (H, W). Each pixel value
                 corresponds to the index of the inferred class. Names for those classes are stored in self.class_names.
        """
        assert image.ndim == 3, f'Image must be in format (H, W, 3). Got {image.ndim} dimensions.'
        original_h, original_w = image.shape[:2]
        image = self.preprocess_image(image=image, input_is_bgr=is_bgr)

        prediction = self.model(image[None,...])[0]
        prediction = self.postprocess_prediction(prediction=prediction, original_h=original_h, original_w=original_w,
                                                 prediction_as_tensor=prediction_as_tensor)
        if return_as_dict:
            prediction = self.build_prediction_dict(prediction=prediction)
        return prediction

    def preprocess_image(self, image: np.ndarray | torch.Tensor, input_is_bgr = False) -> torch.Tensor:
        """
        Preprocess the image to be used for inference. It means to transform it to a tensor, normalize it,
        and convert it to the right model input format (C, H, W) with channels in RGB. Ensuring that the image
        size is a multiple of the model's scale factor.

        :param image: np.ndarray or torch.tensor as uint8 (0, 255) or float32 (0., 1.). The image to segment,
                      in format (H, W, C). Channels must be 3, RGB or BGR.
        :param input_is_bgr: bool. If True, the image is in BGR format. If False, the image is in RGB format.
        :return: torch.Tensor in format (C, H, W). The image in the right format for the model.
        """
        assert image.ndim == 3, f'Image must be 3D. Got {image.shape}.'
        assert image.shape[-1] == 3, f'Image must have 3 channels (Format (H, W, C)). Got {image.shape}.'

        # Convert the image to the right format. Making sure it is a Tensor
        image = image_to_float_tensor(image=image, is_bgr=input_is_bgr, device=self.inference_device)

        # Normalize the image
        image = (image - self.mean) / self.std

        # Convert from HWC to CHW
        h, w, _ = image.shape
        image = image.permute((2, 0, 1))

        # Pad it if needed to make sure it is multiple of scale_factor
        h_scale_diff, w_scale_diff = h % self.scale_factor, w % self.scale_factor
        if h_scale_diff != 0 or w_scale_diff != 0:
            (r_w, l_w), (r_h, l_h) = get_l_r_pad(value=w_scale_diff), get_l_r_pad(value=h_scale_diff)
            image = F.pad(input=image, pad=(l_w, r_w, l_h, r_h), mode='symmetric')

        return image

    def postprocess_prediction(self, prediction: torch.Tensor, original_h: int, original_w: int,
                               prediction_as_tensor: bool = False) -> np.ndarray:
        """
        Postprocess the prediction to get the segmentation map.
        :param prediction: torch.Tensor of shape (N_CLASSES, H, W). The raw predictions of the model.
        :param original_h: int. The original height of the image.
        :param original_w: int. The original width of the image.
        :param prediction_as_tensor: bool. If True, the output is a torch.tensor. If False, the output is a np.ndarray.

        :return: np.ndarray or torch.Tensor of shape (H, W) with type uint8. Inside a dict with the segmenter
                 meta-information if return_as_dict is True. The segmentation map, with values in the range
                 [0, N_CLASSES-1]. Each pixel value corresponds to the index of the inferred class.
        """
        assert prediction.ndim == 3, f'Prediction must be 3D. Got {prediction.shape}.'
        assert prediction.shape[0] == self.num_classes, f'Prediction must have {self.num_classes} channels. ' \
                                                        f'Got {prediction.shape}.'

        prediction = F.interpolate(input=prediction[None], scale_factor=self.scale_factor,
                                   mode='bilinear', align_corners=True)[0]
        pred_h, pred_w = prediction.shape[1:]

        if pred_h != original_h or pred_w != original_w:
            assert pred_h >= original_h and pred_w >= original_w, \
                f'Prediction must be at least as large as the original image. ' \
                f'Got {pred_h}x{pred_w} instead of {original_h}x{original_w}.'
            (l_h, r_h), (l_w, r_w) = get_l_r_pad(value=pred_h - original_h), get_l_r_pad(value=pred_w - original_w)
            prediction = prediction[:, l_h:pred_h-r_h, l_w:pred_w-r_w]

        prediction = prediction.argmax(dim=0).type(torch.uint8)

        if prediction_as_tensor:
            return prediction

        if self.inference_device.type.startswith('cuda'):
            prediction = prediction.cpu()
        prediction = prediction.numpy()
        assert prediction.dtype == np.uint8, f'Prediction must be uint8. Got {prediction.dtype}.'
        return prediction

    def build_prediction_dict(self, prediction: np.ndarray) -> dict:
        """
        Build a dictionary with the prediction, containing the segmentation map and the class name correspondences
        so that the segmenter meta information can be directly used

        :param prediction: np.ndarray of shape (H, W). The segmentation map, with values in the range [0, N_CLASSES-1].
        :return: dict. The dictionary with the prediction.
        """

        assert prediction.ndim == 2, f'Prediction must be 2D. Got {prediction.shape}.'
        assert prediction.dtype == np.uint8, f'Prediction should be uint8. Got {prediction.dtype}.'

        prediction_dict = {SEGMENTATION_MASK: prediction,
                           CLASS_NAMES_LIST: self.class_names,
                           BACKGROUND_CLASS_IDX: self.background_class_idx}

        return prediction_dict



# coding=utf-8
# Description:  visualize yolo label image.

import argparse
import os
import cv2
import numpy as np

IMG_FORMATS = ["bmp", "jpg", "jpeg", "png", "tif", "tiff", "dng", "webp", "mpo"]


def main(args):
    img_dir, label_dir, class_names = args.img_dir, args.label_dir, args.class_names

    label_map = dict()
    for class_id, classname in enumerate(class_names):
        label_map[class_id] = classname

    for file in os.listdir(img_dir):
        if file.split('.')[-1] not in IMG_FORMATS:
            print(f'[Warning]: Non-image file {file}')
            continue
        img_path = os.path.join(img_dir, file)
        label_path = os.path.join(label_dir, file[: file.rindex('.')] + '.txt')

        try:
            img_data = cv2.imread(img_path)
            height, width, _ = img_data.shape
            color = [tuple(np.random.choice(range(256), size=3)) for i in class_names]
            thickness = 2

            with open(label_path, 'r') as f:
                for bbox in f:
                    cls, x_c, y_c, w, h = [float(v) if i > 0 else int(v) for i, v in enumerate(bbox.split('\n')[0].split(' '))]

                    x_tl = int((x_c - w / 2) * width)
                    y_tl = int((y_c - h / 2) * height)
                    cv2.rectangle(img_data, (x_tl, y_tl), (x_tl + int(w * width), y_tl + int(h * height)), tuple([int(x) for x in color[cls]]), thickness)
                    cv2.putText(img_data, label_map[cls], (x_tl, y_tl - 10), cv2.FONT_HERSHEY_COMPLEX, 1, tuple([int(x) for x in color[cls]]), thickness)

            cv2.imshow('image', img_data)
            cv2.waitKey(0)
        except Exception as e:
            print(f'[Error]: {e} {img_path}')
    print('======All Done!======')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--img_dir', default='VOCdevkit/voc_07_12/images')
    parser.add_argument('--label_dir', default='VOCdevkit/voc_07_12/labels')
    parser.add_argument('--class_names', default=['aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus', 'car', 'cat', 'chair', 'cow', 'diningtable', 'dog',
                        'horse', 'motorbike', 'person', 'pottedplant', 'sheep', 'sofa', 'train', 'tvmonitor'])

    args = parser.parse_args()
    print(args)

    main(args)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# Copyright (c) 2014-2021 Megvii Inc. All rights reserved.

import torch
import torch.distributed as dist
import torch.nn as nn

import os
import random

from .base_exp import BaseExp


class Exp(BaseExp):
    def __init__(self):
        super().__init__()

        # ---------------- model config ---------------- #
        self.num_classes = 80
        self.depth = 1.00
        self.width = 1.00

        # ---------------- dataloader config ---------------- #
        # set worker to 4 for shorter dataloader init time
        self.data_num_workers = 4
        self.input_size = (640, 640)
        self.random_size = (14, 26)
        self.train_ann = "instances_train2017.json"
        self.val_ann = "instances_val2017.json"

        # --------------- transform config ----------------- #
        self.degrees = 10.0
        self.translate = 0.1
        self.scale = (0.1, 2)
        self.mscale = (0.8, 1.6)
        self.shear = 2.0
        self.perspective = 0.0
        self.enable_mixup = True

        # --------------  training config --------------------- #
        self.warmup_epochs = 5
        self.max_epoch = 300
        self.warmup_lr = 0
        self.basic_lr_per_img = 0.01 / 64.0
        self.scheduler = "yoloxwarmcos"
        self.no_aug_epochs = 15
        self.min_lr_ratio = 0.05
        self.ema = True

        self.weight_decay = 5e-4
        self.momentum = 0.9
        self.print_interval = 10
        self.eval_interval = 10
        self.exp_name = os.path.split(os.path.realpath(__file__))[1].split(".")[0]

        # -----------------  testing config ------------------ #
        self.test_size = (640, 640)
        self.test_conf = 0.001
        self.nmsthre = 0.65

    def get_model(self):
        from yolox.models import YOLOPAFPN, YOLOX, YOLOXHead

        def init_yolo(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03

        if getattr(self, "model", None) is None:
            in_channels = [256, 512, 1024]
            backbone = YOLOPAFPN(self.depth, self.width, in_channels=in_channels)
            head = YOLOXHead(self.num_classes, self.width, in_channels=in_channels)
            self.model = YOLOX(backbone, head)

        self.model.apply(init_yolo)
        self.model.head.initialize_biases(1e-2)
        return self.model

    def get_data_loader(self, batch_size, is_distributed, no_aug=False):
        from yolox.data import (
            COCODataset,
            DataLoader,
            InfiniteSampler,
            MosaicDetection,
            TrainTransform,
            YoloBatchSampler
        )

        dataset = COCODataset(
            data_dir=None,
            json_file=self.train_ann,
            img_size=self.input_size,
            preproc=TrainTransform(
                rgb_means=(0.485, 0.456, 0.406),
                std=(0.229, 0.224, 0.225),
                max_labels=50,
            ),
        )

        dataset = MosaicDetection(
            dataset,
            mosaic=not no_aug,
            img_size=self.input_size,
            preproc=TrainTransform(
                rgb_means=(0.485, 0.456, 0.406),
                std=(0.229, 0.224, 0.225),
                max_labels=120,
            ),
            degrees=self.degrees,
            translate=self.translate,
            scale=self.scale,
            shear=self.shear,
            perspective=self.perspective,
            enable_mixup=self.enable_mixup,
        )

        self.dataset = dataset

        if is_distributed:
            batch_size = batch_size // dist.get_world_size()

        sampler = InfiniteSampler(len(self.dataset), seed=self.seed if self.seed else 0)

        batch_sampler = YoloBatchSampler(
            sampler=sampler,
            batch_size=batch_size,
            drop_last=False,
            input_dimension=self.input_size,
            mosaic=not no_aug,
        )

        dataloader_kwargs = {"num_workers": self.data_num_workers, "pin_memory": True}
        dataloader_kwargs["batch_sampler"] = batch_sampler
        train_loader = DataLoader(self.dataset, **dataloader_kwargs)

        return train_loader

    def random_resize(self, data_loader, epoch, rank, is_distributed):
        tensor = torch.LongTensor(2).cuda()

        if rank == 0:
            size_factor = self.input_size[1] * 1.0 / self.input_size[0]
            size = random.randint(*self.random_size)
            size = (int(32 * size), 32 * int(size * size_factor))
            tensor[0] = size[0]
            tensor[1] = size[1]

        if is_distributed:
            dist.barrier()
            dist.broadcast(tensor, 0)

        input_size = data_loader.change_input_dim(
            multiple=(tensor[0].item(), tensor[1].item()), random_range=None
        )
        return input_size

    def get_optimizer(self, batch_size):
        if "optimizer" not in self.__dict__:
            if self.warmup_epochs > 0:
                lr = self.warmup_lr
            else:
                lr = self.basic_lr_per_img * batch_size

            pg0, pg1, pg2 = [], [], []  # optimizer parameter groups

            for k, v in self.model.named_modules():
                if hasattr(v, "bias") and isinstance(v.bias, nn.Parameter):
                    pg2.append(v.bias)  # biases
                if isinstance(v, nn.BatchNorm2d) or "bn" in k:
                    pg0.append(v.weight)  # no decay
                elif hasattr(v, "weight") and isinstance(v.weight, nn.Parameter):
                    pg1.append(v.weight)  # apply decay

            optimizer = torch.optim.SGD(
                pg0, lr=lr, momentum=self.momentum, nesterov=True
            )
            optimizer.add_param_group(
                {"params": pg1, "weight_decay": self.weight_decay}
            )  # add pg1 with weight_decay
            optimizer.add_param_group({"params": pg2})
            self.optimizer = optimizer

        return self.optimizer

    def get_lr_scheduler(self, lr, iters_per_epoch):
        from yolox.utils import LRScheduler

        scheduler = LRScheduler(
            self.scheduler,
            lr,
            iters_per_epoch,
            self.max_epoch,
            warmup_epochs=self.warmup_epochs,
            warmup_lr_start=self.warmup_lr,
            no_aug_epochs=self.no_aug_epochs,
            min_lr_ratio=self.min_lr_ratio,
        )
        return scheduler

    def get_eval_loader(self, batch_size, is_distributed, testdev=False):
        from yolox.data import COCODataset, ValTransform

        valdataset = COCODataset(
            data_dir=None,
            json_file=self.val_ann if not testdev else "image_info_test-dev2017.json",
            name="val2017" if not testdev else "test2017",
            img_size=self.test_size,
            preproc=ValTransform(
                rgb_means=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)
            ),
        )

        if is_distributed:
            batch_size = batch_size // dist.get_world_size()
            sampler = torch.utils.data.distributed.DistributedSampler(
                valdataset, shuffle=False
            )
        else:
            sampler = torch.utils.data.SequentialSampler(valdataset)

        dataloader_kwargs = {
            "num_workers": self.data_num_workers,
            "pin_memory": True,
            "sampler": sampler,
        }
        dataloader_kwargs["batch_size"] = batch_size
        val_loader = torch.utils.data.DataLoader(valdataset, **dataloader_kwargs)

        return val_loader

    def get_evaluator(self, batch_size, is_distributed, testdev=False):
        from yolox.evaluators import COCOEvaluator

        val_loader = self.get_eval_loader(batch_size, is_distributed, testdev=testdev)
        evaluator = COCOEvaluator(
            dataloader=val_loader,
            img_size=self.test_size,
            confthre=self.test_conf,
            nmsthre=self.nmsthre,
            num_classes=self.num_classes,
            testdev=testdev,
        )
        return evaluator

    def eval(self, model, evaluator, is_distributed, half=False):
        return evaluator.evaluate(model, is_distributed, half)



import os
import numpy as np
prefix_dir = 'MOT17/'
root_dir = 'train/'
result_csv = 'train_half_annots.csv'
train_half_set = {2: 301, 4: 526, 5:419, 9:263, 10:328, 11:451, 13:376}
fout = open(result_csv, 'w')

for data_name in sorted(os.listdir(prefix_dir + root_dir)):
	print(data_name)
	gt_path = os.path.join(prefix_dir, root_dir, data_name, 'gt', 'gt.txt')
	# print(gt_path)
	data_raw = np.loadtxt(gt_path, delimiter=',', dtype='float', usecols=(0,1,2,3,4,5,6,7,8))

	data_sort = data_raw[np.lexsort(data_raw[:,::-1].T)]
	visible_raw = data_sort[:,8]
	# print(data_sort)
	# print(data_sort[-1, 0])
	img_num = data_sort[-1, 0]

	# print(data_sort.shape[0])
	box_num = data_sort.shape[0]

	person_box_num = np.sum(data_sort[:,6] == 1)
	# print(person_box_num)
# 	import ipdb; ipdb.set_trace()
	for i in range(box_num):
		c = int(data_sort[i, 6])
		v = visible_raw[i]
		img_index = int(data_sort[i, 0])
		if c == 1 and v > 0.1 and img_index < train_half_set[int(data_name[-2:])]:
			img_index = int(data_sort[i, 0])
			img_name = data_name + '/img1/' + str(img_index).zfill(6) + '.jpg'
			print(root_dir + img_name + ', ' + str(int(data_sort[i, 1])) + ', ' + str(data_sort[i, 2]) + ', ' + str(data_sort[i, 3]) + ', ' + str(data_sort[i, 2] + data_sort[i, 4]) + ', ' + str(data_sort[i, 3] + data_sort[i, 5]) + ', person\n')
			fout.write(root_dir + img_name + ', ' + str(int(data_sort[i, 1])) + ', ' + str(data_sort[i, 2]) + ', ' + str(data_sort[i, 3]) + ', ' + str(data_sort[i, 2] + data_sort[i, 4]) + ', ' + str(data_sort[i, 3] + data_sort[i, 5]) + ', person\n')

fout.close()



from __future__ import annotations

import numpy as np
from ultralytics.yolo.engine.results import Results
from time import time
from loguru import logger

from vid2info.inference.config import POLYGON_XY, POLYGON_XYN, CLASS_ID, CLASS_NAME, CONFIDENCE, TRACK_INFO, TRACK_ID, BBOX_INFO, \
    XNMIN, YNMIN, XNMAX, YNMAX, XMIN, YMIN, XMAX, YMAX, BBOX_XYXY, BBOX_XYXYN, SEGMENTATION_INFO, IMAGE_HW, TIMESTAMP, \
    BBOX_WH, BBOX_WHN, BBOX_CXCY, BBOX_CXCYN


def yolov8_results_to_info_dict(detections: Results,
                                avoid_non_tracked_objects: bool = True) -> \
        dict[int, dict[str, int | float | dict[str, int | float | np.ndarray] | tuple[int, int] | None]]:
    """
    Convert the detections format of YOLOv8 to a dictionary. This cast is useful to separate our logical structure
    from the YOLOv8 implementation. This way, we could easily change the model without affecting the rest of the code.

    :param detections: The detections returned by the YOLOv8 model.
    :return: A dictionary where the keys are track ids and the values are
                another dictionary with detailed detection info.
    """

    assert avoid_non_tracked_objects, 'Only avoid_non_tracked_objects=True is supported at the moment.'
    tracks = {}

    for detection in detections:
        boxes = detection.boxes.numpy()
        if not boxes.is_track:
            logger.warning(f'Found a non-track box. Skipping it.')
            continue

        # Validate the box format
        assert len(boxes.xyxyn) == 1, f'Expected only one xyxyn bbox per detection. Got {len(boxes.xyxyn)}.'
        assert len(boxes.xyxy) == 1, f'Expected only one xyxy bbox per detection. Got {len(boxes.xyxy)}.'
        xn_min, yn_min, xn_max, yn_max = map(float, boxes.xyxyn[0])
        x_min, y_min, x_max, y_max = map(float, boxes.xyxy[0])

        track_id, class_id = int(boxes.id), int(boxes.cls)
        confidence = float(boxes.conf)

        # Initialize segmentation_info to None, update it if masks are present
        segmentation_info = None
        if 'masks' in detection.keys:
            # Validate the mask format
            assert len(detection.masks.xy) == 1, f'Expected only one xy polygon per detection. Got {len(detection.masks.xy)}.'
            assert len(detection.masks.xyn) == 1, f'Expected only one xyn polygon per detection. Got {len(detection.masks.xyn)}.'

            # Masks won't be saved, only polygons, to avoid memory issues
            segmentation_info = {POLYGON_XY: detection.masks.xy[0],
                                 POLYGON_XYN: detection.masks.xyn[0]}

        # Store all the information in the tracks dictionary
        tracks[track_id] = {
            CLASS_ID: class_id,
            CLASS_NAME: detection.names[class_id],
            CONFIDENCE: confidence,
            TRACK_INFO: {
                TRACK_ID: track_id,
            },
            BBOX_INFO: {
                XNMIN: xn_min, YNMIN: yn_min, XNMAX: xn_max, YNMAX: yn_max,
                XMIN: x_min, YMIN: y_min, XMAX: x_max, YMAX: y_max,
                BBOX_XYXY: boxes.xyxy[0],
                BBOX_XYXYN: boxes.xyxyn[0],
                BBOX_WH: boxes.xywh[0, 2:],
                BBOX_WHN: boxes.xywhn[0, 2:],
                BBOX_CXCY: boxes.xywh[0, :2],
                BBOX_CXCYN: boxes.xywhn[0, :2],
            },
            SEGMENTATION_INFO: segmentation_info,
            IMAGE_HW: detection.orig_shape,
            TIMESTAMP: time(),
        }

    return tracks



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# Code are based on
# https://github.com/facebookresearch/detectron2/blob/master/detectron2/engine/launch.py
# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) Megvii, Inc. and its affiliates.

from loguru import logger

import torch
import torch.distributed as dist
import torch.multiprocessing as mp

import yolox.utils.dist as comm
from yolox.utils import configure_nccl

import os
import subprocess
import sys
import time

__all__ = ["launch"]


def _find_free_port():
    """
    Find an available port of current machine / node.
    """
    import socket

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # Binding to port 0 will cause the OS to find an available port for us
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    # NOTE: there is still a chance the port could be taken by other processes.
    return port


def launch(
    main_func,
    num_gpus_per_machine,
    num_machines=1,
    machine_rank=0,
    backend="nccl",
    dist_url=None,
    args=(),
):
    """
    Args:
        main_func: a function that will be called by `main_func(*args)`
        num_machines (int): the total number of machines
        machine_rank (int): the rank of this machine (one per machine)
        dist_url (str): url to connect to for distributed training, including protocol
                       e.g. "tcp://127.0.0.1:8686".
                       Can be set to auto to automatically select a free port on localhost
        args (tuple): arguments passed to main_func
    """
    world_size = num_machines * num_gpus_per_machine
    if world_size > 1:
        if int(os.environ.get("WORLD_SIZE", "1")) > 1:
            dist_url = "{}:{}".format(
                os.environ.get("MASTER_ADDR", None),
                os.environ.get("MASTER_PORT", "None"),
            )
            local_rank = int(os.environ.get("LOCAL_RANK", "0"))
            world_size = int(os.environ.get("WORLD_SIZE", "1"))
            _distributed_worker(
                local_rank,
                main_func,
                world_size,
                num_gpus_per_machine,
                num_machines,
                machine_rank,
                backend,
                dist_url,
                args,
            )
            exit()
        launch_by_subprocess(
            sys.argv,
            world_size,
            num_machines,
            machine_rank,
            num_gpus_per_machine,
            dist_url,
            args,
        )
    else:
        main_func(*args)


def launch_by_subprocess(
    raw_argv,
    world_size,
    num_machines,
    machine_rank,
    num_gpus_per_machine,
    dist_url,
    args,
):
    assert (
        world_size > 1
    ), "subprocess mode doesn't support single GPU, use spawn mode instead"

    if dist_url is None:
        # ------------------------hack for multi-machine training -------------------- #
        if num_machines > 1:
            master_ip = subprocess.check_output(["hostname", "--fqdn"]).decode("utf-8")
            master_ip = str(master_ip).strip()
            dist_url = "tcp://{}".format(master_ip)
            ip_add_file = "./" + args[1].experiment_name + "_ip_add.txt"
            if machine_rank == 0:
                port = _find_free_port()
                with open(ip_add_file, "w") as ip_add:
                    ip_add.write(dist_url+'\n')
                    ip_add.write(str(port))
            else:
                while not os.path.exists(ip_add_file):
                    time.sleep(0.5)

                with open(ip_add_file, "r") as ip_add:
                    dist_url = ip_add.readline().strip()
                    port = ip_add.readline()
        else:
            dist_url = "tcp://127.0.0.1"
            port = _find_free_port()

    # set PyTorch distributed related environmental variables
    current_env = os.environ.copy()
    current_env["MASTER_ADDR"] = dist_url
    current_env["MASTER_PORT"] = str(port)
    current_env["WORLD_SIZE"] = str(world_size)
    assert num_gpus_per_machine <= torch.cuda.device_count()

    if "OMP_NUM_THREADS" not in os.environ and num_gpus_per_machine > 1:
        current_env["OMP_NUM_THREADS"] = str(1)
        logger.info(
            "\n*****************************************\n"
            "Setting OMP_NUM_THREADS environment variable for each process "
            "to be {} in default, to avoid your system being overloaded, "
            "please further tune the variable for optimal performance in "
            "your application as needed. \n"
            "*****************************************".format(
                current_env["OMP_NUM_THREADS"]
            )
        )

    processes = []
    for local_rank in range(0, num_gpus_per_machine):
        # each process's rank
        dist_rank = machine_rank * num_gpus_per_machine + local_rank
        current_env["RANK"] = str(dist_rank)
        current_env["LOCAL_RANK"] = str(local_rank)

        # spawn the processes
        cmd = ["python3", *raw_argv]

        process = subprocess.Popen(cmd, env=current_env)
        processes.append(process)

    for process in processes:
        process.wait()
        if process.returncode != 0:
            raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd)


def _distributed_worker(
    local_rank,
    main_func,
    world_size,
    num_gpus_per_machine,
    num_machines,
    machine_rank,
    backend,
    dist_url,
    args,
):
    assert (
        torch.cuda.is_available()
    ), "cuda is not available. Please check your installation."
    configure_nccl()
    global_rank = machine_rank * num_gpus_per_machine + local_rank
    logger.info("Rank {} initialization finished.".format(global_rank))
    try:
        dist.init_process_group(
            backend=backend,
            init_method=dist_url,
            world_size=world_size,
            rank=global_rank,
        )
    except Exception:
        logger.error("Process group URL: {}".format(dist_url))
        raise
    # synchronize is needed here to prevent a possible timeout after calling init_process_group
    # See: https://github.com/facebookresearch/maskrcnn-benchmark/issues/172
    comm.synchronize()

    if global_rank == 0 and os.path.exists(
        "./" + args[1].experiment_name + "_ip_add.txt"
    ):
        os.remove("./" + args[1].experiment_name + "_ip_add.txt")

    assert num_gpus_per_machine <= torch.cuda.device_count()
    torch.cuda.set_device(local_rank)

    args[1].local_rank = local_rank
    args[1].num_machines = num_machines

    # Setup the local process group (which contains ranks within the same machine)
    # assert comm._LOCAL_PROCESS_GROUP is None
    # num_machines = world_size // num_gpus_per_machine
    # for i in range(num_machines):
    # ranks_on_i = list(range(i * num_gpus_per_machine, (i + 1) * num_gpus_per_machine))
    # pg = dist.new_group(ranks_on_i)
    # if i == machine_rank:
    # comm._LOCAL_PROCESS_GROUP = pg

    main_func(*args)
